@@ -1,30 +1,35 @@
 """check_arcce_monitor - Fetches ARC jobs and posts results."""
 
+import asyncio
 import os
 import random
 import re
 import time
+from typing import Optional, Tuple
 from arcnagios import arcutils, nagutils, utils
-from arcnagios.nagutils import ServiceReport, ServiceOk, ServiceUnknown, \
-                               OK, WARNING, CRITICAL
-from arcnagios.arcutils import ArcgetProcess, ParseError
-from arcnagios.utils import nth, counted_adjectives
-from arcnagios.persistence import ObsoletePersistentObject
-from arcnagios.ce.jobutils import JobNagiosPlugin
+from arcnagios.arcutils import Arcstat
+from arcnagios.arcclients import arcget, arcstat
+from arcnagios.nagutils import \
+        ServiceReport, ServiceOk, ServiceUnknown, OK, WARNING, CRITICAL, \
+        NagiosPerflogTimer
+from arcnagios.utils import nth, counted_adjectives, host_of_uri
+from arcnagios.ce.jobutils import JobInfo, JobNagiosPlugin
 
-_arcget_harmless_re = re.compile('|'.join([
+_SPECIFIC_STATE_TR = dict((ord(c), None) for c in "'=[],:")
+
+_ARCGET_HARMLESS_RE = re.compile('|'.join([
     r'$',
     r'Results stored at:.*',
     r'Warning: Some jobs were not removed.*',
 ]))
 
-def check_arcget_output(output):
+def check_arcget_output(output: str) -> Tuple[bool, bool]:
     need_arcclean = False
-    for ln in output.split('\n'):
-        ln = ln.strip()
-        if ln.startswith('Use arcclean to remove'):
+    for line in output.split('\n'):
+        line = line.strip()
+        if line.startswith('Use arcclean to remove'):
             need_arcclean = True
-        elif not re.match(_arcget_harmless_re, ln):
+        elif not re.match(_ARCGET_HARMLESS_RE, line):
             return False, need_arcclean
     return True, need_arcclean
 
@@ -39,7 +44,7 @@ class JobActionStats:
         self.retry_count = 0
         self.postponed_count = 0
 
-    def load(self):
+    def load(self) -> float:
         return self.postponed_count \
              / float(self.discarded_count + self.cleaned_count + 1)
 
@@ -47,71 +52,76 @@ class Check_arcce_monitor(JobNagiosPlugin):
     def __init__(self):
         JobNagiosPlugin.__init__(self)
         self.start_time = time.time()
-        ap = self.argparser
-        ap.add_argument('--ce', dest = 'ces',
+        argp = self.argparser
+        argp.add_argument('--ce', dest = 'ces',
                 default = [], action = 'append',
                 metavar = 'CE',
                 help = 'Pass one or more times to restrict monitoring '
                        'to the given CEs.')
-        ap.add_argument('--job-tag', dest = 'job_tags',
+        argp.add_argument('--job-tag', dest = 'job_tags',
                 default = [], action = 'append',
                 metavar = 'TAG',
                 help = 'Pass one or more times to restrict monitoring '
                        'to the given job tags. Use "default" to match '
                        'jobs started without a --job-tag option.')
-        ap.add_argument('--termination-service', dest = 'termination_service',
+        argp.add_argument('--termination-service', dest = 'termination_service',
                 default = 'ARCCE Job Termination',
                 help = 'Default service to submit result to if not specified '
                        'when submitting the job. '
                        'Deprecated: Should be passed on submission.')
-        ap.add_argument('--max-sysinfo-lag', dest = 'max_infosys_lag',
+        argp.add_argument('--max-sysinfo-lag', dest = 'max_infosys_lag',
                 default = 3600.0, metavar = 'T',
                 help = 'The maximum time to wait for a job to turn up in '
                        'the arcstat listing before abandoning it.')
-        ap.add_argument('--max-check-attempts', dest = 'max_check_attempts',
+        argp.add_argument('--max-check-attempts', dest = 'max_check_attempts',
                 default = 12, metavar = 'N',
                 help = 'The maximum number of consecutive times a job in '
                        'post-SUBMITTED state is absent from arcstat listing '
                        'before it is abandoned.')
-        ap.add_argument('--max-fetch-attempts', dest = 'max_fetch_attempts',
+        argp.add_argument('--max-fetch-attempts', dest = 'max_fetch_attempts',
                 default = 8, metavar = 'N',
                 help = 'The maximum number of attempts to fetch a job before '
                        'abandoning it.')
-        ap.add_argument('--keep-failed-jobdata', dest = 'keep_failed_jobdata',
+        argp.add_argument('--keep-failed-jobdata', dest = 'keep_failed_jobdata',
                 action = 'store_true', default = False,
                 help = 'Keep the job descriptions and output directories for '
                        'failed jobs. These will not be removed automatically.')
-        ap.add_argument('--keep-all-jobdata', dest = 'keep_all_jobdata',
+        argp.add_argument('--keep-all-jobdata', dest = 'keep_all_jobdata',
                 action = 'store_true', default = False,
                 help = 'Keep the job descriptions and output directories for '
                        'all jobs. These will not be removed automatically.')
-        ap.add_argument('--timeout', dest = 'timeout',
+        argp.add_argument('--timeout', dest = 'timeout',
                 type = int, default = 110,
                 help = 'Approximate overall timeout.')
-        ap.add_argument('--arcstat-timeout', dest = 'arcstat_timeout',
+        argp.add_argument('--arcstat-timeout', dest = 'arcstat_timeout',
                 type = int, default = 5, metavar = 'T',
                 help = 'Passed to arcstat --timeout.')
-        ap.add_argument('--arcget-timeout', dest = 'arcget_timeout',
+        argp.add_argument('--arcget-timeout', dest = 'arcget_timeout',
                 type = int, default = 10,
                 help = 'Maximum value to pass to arcget --timeout.')
-        ap.add_argument('-c', dest = 'critical_load',
+        argp.add_argument('--show-arcget-output-here',
+                dest = 'show_arcget_output_here',
+                action = 'store_true', default = False,
+                help = 'Include output from failed arcget in the this service '
+                       'in addition to posting it to the termination service.')
+        argp.add_argument('-c', dest = 'critical_load',
                 type = float, default = 20,
                 help = 'Ratio of remaining work to processed work above which \
                         to issue a critical alert.')
-        ap.add_argument('-w', dest = 'warning_load',
+        argp.add_argument('-w', dest = 'warning_load',
                 type = float, default = 10,
                 help = 'Ratio of remaining work to processed work above which \
                         to issue a warning alert.')
 
-    def time_left(self):
+    def time_left(self) -> float:
         return self.opts.timeout - time.time() + self.start_time
 
-    def parse_args(self, args = None):
+    def parse_args(self, args) -> None:
         """Parse ARCCE-specific command-line options."""
 
-        JobNagiosPlugin.parse_args(self, args = None)
+        JobNagiosPlugin.parse_args(self, args)
 
-    def _clean_output_dir(self, output_dir):
+    def _clean_output_dir(self, output_dir: str) -> None:
         conflict = '.conflict-%d' % int(time.time())
         for entry in os.listdir(output_dir):
             if not '.conflict-' in entry:
@@ -120,20 +130,20 @@ class Check_arcce_monitor(JobNagiosPlugin):
                                  subdir)
                 os.rename(subdir, subdir + conflict)
 
-    def prepare_top_output_dir(self, jobinfo):
+    def prepare_top_output_dir(self, jobinfo: JobInfo) -> str:
         workdir = self.workdir_for(jobinfo.host, jobinfo.job_tag)
         job_output_dir = os.path.join(workdir, self.JOB_OUTPUT_DIRNAME)
         if os.path.exists(job_output_dir):
             self._clean_output_dir(job_output_dir)
         return job_output_dir
 
-    def locate_output_dir(self, top_output_dir):
+    def locate_output_dir(self, top_output_dir: str) -> Optional[str]:
         for subdir in os.listdir(top_output_dir):
             if not subdir in ['.', '..'] or '.conflict-' in subdir:
                 return os.path.join(top_output_dir, subdir)
         return None
 
-    def fetch_job(self, jobinfo, job_error = None):
+    async def fetch_job(self, jobinfo: JobInfo, job_error: Optional[str] =None):
         """Fetch the job described by `jobinfo : JobInfo`, submit passive
         results, and return a tuple `(did_fetch, check_ok, status_code)`, where
 
@@ -161,13 +171,13 @@ class Check_arcce_monitor(JobNagiosPlugin):
         self.log.info('Fetching job %s in terminal state %s.',
                       jobinfo.job_id, jobinfo.job_state)
         top_output_dir = self.prepare_top_output_dir(jobinfo)
-        arcget_process = \
-            ArcgetProcess(
-                jobinfo.job_id, top_output_dir = top_output_dir,
-                timeout = min(self.opts.arcget_timeout, self.time_left()),
-                perflog = self.perflog)
-        yield
-        arcget_result = arcget_process.communicate()
+        async with self._arcclients_semaphore:
+            with NagiosPerflogTimer(
+                    self.perflog, "arcget_time", host_of_uri(jobinfo.job_id)):
+                timeout = min(self.opts.arcget_timeout, self.time_left())
+                arcget_result = await arcget(
+                        jobinfo.job_id, top_output_dir = top_output_dir,
+                        timeout = timeout, log = self.log)
         job_output_dir = self.locate_output_dir(top_output_dir)
         if job_output_dir is None:
             if arcget_result.is_ok():
@@ -183,24 +193,24 @@ class Check_arcce_monitor(JobNagiosPlugin):
                 self.log.error('Failed to fetch %s.', jobinfo.job_id)
                 termination_report.update_status(nagutils.WARNING,
                         'Failed to fetch job.')
-                if arcget_result.error.output:
+                if arcget_result.get_error().output:
                     details = 'Output from arcget:\n%s' \
-                            % arcget_result.error.output
-                    self.log.error(details)
+                            % arcget_result.get_error().output
+                    if self.opts.show_argget_output_here:
+                        self.log.error(details)
                     termination_report.log.error(details)
                 termination_report.log.error('JID: %s', jobinfo.job_id)
                 did_fetch = False
                 ok_check = True
-            yield (did_fetch, ok_check, termination_report.status_code)
-            return
+            return (did_fetch, ok_check, termination_report.status_code)
 
         # Check if arcget returned non-zero despite having fetched something.
         if arcget_result.is_error():
             is_harmless, need_arcclean \
-                    = check_arcget_output(arcget_result.error.output)
+                    = check_arcget_output(arcget_result.get_error().output)
             if need_arcclean:
                 termination_report.log.warning('Separate arcclean needed.')
-                self.cleaner.call('arcclean', jobinfo.job_id)
+                await self.cleaner.call('arcclean', jobinfo.job_id)
             if not is_harmless:
                 termination_report.update_status(nagutils.WARNING,
                     '%s: %s' % (jobinfo.job_id, arcget_result))
@@ -210,13 +220,12 @@ class Check_arcce_monitor(JobNagiosPlugin):
                 utils.file_contents(os.path.join(job_output_dir, 'stderr.txt'))
             if not errors is None and errors.strip() != '':
                 self.log.error('stderr.txt for %s:', jobinfo.job_id)
-                for ln in errors.strip().split('\n'):
-                    self.log.error('.. %s', ln)
+                for line in errors.strip().split('\n'):
+                    self.log.error('.. %s', line)
                 details = 'stderr.txt:\n%s' % errors
                 termination_report.log.error(details)
             termination_report.log.error('JID: %s', jobinfo.job_id)
-            yield (True, True, termination_report.status_code)
-            return
+            return (True, True, termination_report.status_code)
 
         # Run check and publish results from job tests.
         termination_report.log.info('JID: %s', jobinfo.job_id)
@@ -234,9 +243,9 @@ class Check_arcce_monitor(JobNagiosPlugin):
 
         if status_code != nagutils.OK:
             termination_report.log.error('JID: %s', jobinfo.job_id)
-        yield (True, True, status_code)
+        return (True, True, status_code)
 
-    def check_job_progress(self, jobinfo):
+    def check_job_progress(self, jobinfo: JobInfo) -> None:
         if jobinfo.job_state_time is None or jobinfo.progress_service is None:
             return
         attrs = {}
@@ -267,22 +276,24 @@ class Check_arcce_monitor(JobNagiosPlugin):
             report.update_status(status, msg)
             jobinfo.job_state_alert = status
 
-    def report_job_state_time(self, jobinfo, t_now):
+    def report_job_state_time(self, jobinfo: JobInfo, t_now: float) -> None:
         if not self.opts.granular_perfdata:
             return
+        assert jobinfo.job_state_time
         dt_state = t_now - jobinfo.job_state_time
         perf_indices = [
             'host:%s' % jobinfo.host,
             'arc_internal_job_state:%s' % jobinfo.job_state,
         ]
         if not jobinfo.job_specific_state is None:
-            tr = dict((ord(c), None) for c in "'=[],:")
-            s = str(jobinfo.job_specific_state).translate(tr)
+            s = str(jobinfo.job_specific_state).translate(_SPECIFIC_STATE_TR)
             perf_indices.append('arc_middleware_job_state:%s' % s)
         perf_label = 'arc_job_state_time[%s]' % ','.join(perf_indices)
         self.add_perfdata(perf_label, dt_state, uom = 's', limit_min = 0)
 
-    def process_missing_job(self, jobinfo, job_action_stats):
+    async def process_missing_job(
+            self, jobinfo: JobInfo, job_action_stats: JobActionStats
+        ) -> None:
         # Job missing from from arcstat output can happen
         #   a) right after submission before it becomes available,
         #   b) temporarily if the CE infosys is unavailable, or
@@ -315,10 +326,15 @@ class Check_arcce_monitor(JobNagiosPlugin):
                           'state %s, removing active job info.',
                           jobinfo.job_id, jobinfo.job_tag, jobinfo.host,
                           jobinfo.job_state)
-            self.discard_job(jobinfo, archive = self.opts.keep_failed_jobdata)
+            await self.discard_job(jobinfo,
+                    archive = self.opts.keep_failed_jobdata)
             job_action_stats.discarded_count += 1
 
-    def process_found_job(self, jobinfo, jobstat, job_action_stats):
+    async def process_found_job(
+            self, jobinfo: JobInfo, jobstat: Arcstat,
+            job_action_stats: JobActionStats
+        ) -> None:
+
         jobinfo.check_attempts = 0
         self.log.debug('Checking job on %s.', jobinfo.host)
 
@@ -340,11 +356,8 @@ class Check_arcce_monitor(JobNagiosPlugin):
         elif self.time_left() < 1:
             job_action_stats.postponed_count += 1
         else:
-            result = None
-            for result in self.fetch_job(jobinfo, jobstat.job_error):
-                if result is None:
-                    yield
-            did_fetch, ok_check, passive_status_code = result
+            did_fetch, ok_check, passive_status_code = \
+                await self.fetch_job(jobinfo, jobstat.job_error)
             if not ok_check:
                 job_action_stats.error_count += 1
             archive = self.opts.keep_failed_jobdata \
@@ -358,7 +371,7 @@ class Check_arcce_monitor(JobNagiosPlugin):
                             dist_name, choice_name, ok_rep and 'good' or 'bad')
                     self._reputation_tracker.submit(
                             dist_name, choice_name, ok_rep)
-                self.cleanup_job(jobinfo, archive = archive)
+                await self.cleanup_job(jobinfo, archive = archive)
                 job_action_stats.cleaned_count += 1
             elif (jobinfo.fetch_attempts or 0) < self.opts.max_fetch_attempts:
                 jobinfo.fetch_attempts = (jobinfo.fetch_attempts or 0) + 1
@@ -367,11 +380,10 @@ class Check_arcce_monitor(JobNagiosPlugin):
                 job_action_stats.retry_count += 1
             else:
                 self.log.warning('Giving up on fetching %s.', jobinfo.job_id)
-                self.discard_job(jobinfo, archive = archive)
+                await self.discard_job(jobinfo, archive = archive)
                 job_action_stats.discarded_count += 1
-        yield
 
-    def check(self):
+    async def _check_async(self) -> ServiceReport:
         """Monitor submitted jobs."""
 
         if not os.path.exists(self.top_workdir):
@@ -405,24 +417,15 @@ class Check_arcce_monitor(JobNagiosPlugin):
             if not job_tags is None and not (job_tag or 'default') in job_tags:
                 continue
 
-            workdir = self.workdir_for(host, job_tag)
-            ajf = os.path.join(workdir, self.ACTIVE_JOB_FILENAME)
-            if not os.path.exists(ajf):
+            jobinfo = self.load_active_job(host, job_tag)
+            if jobinfo is None:
                 self.log.debug('No active job info for %s.', host)
+                self.cleanup_job_files(
+                    host, job_tag, archive=self.opts.keep_failed_jobdata)
             else:
-                try:
-                    jobinfo = self.load_active_job(host, job_tag)
-                    jobinfo.host = host
-                    jobinfo.job_tag = job_tag
-                    jobinfo_by_id[jobinfo.job_id] = jobinfo
-                except ObsoletePersistentObject as exn:
-                    self.log.warning(
-                            'Removing outdated job files for %s, tag %s.',
-                            host, job_tag)
-                    self.cleanup_job_files(
-                        host, job_tag, archive=self.opts.keep_failed_jobdata)
-                except (OSError, ParseError) as exn:
-                    self.log.error('Cannot load job file %s: %s', ajf, exn)
+                jobinfo.host = host
+                jobinfo.job_tag = job_tag
+                jobinfo_by_id[jobinfo.job_id] = jobinfo
 
         query_jobids = [jobinfo.job_id for jobinfo in jobinfo_by_id.values()]
         random.shuffle(query_jobids)
@@ -433,30 +436,29 @@ class Check_arcce_monitor(JobNagiosPlugin):
 
         # Obtains information from CEs about the active jobs.
         self.log.debug('Querying job IDs %s', ', '.join(query_jobids))
-        try:
-            jobstat_by_id = arcutils.arcstat(query_jobids, log = self.log,
-                                             timeout = self.opts.arcstat_timeout)
-        except ParseError as exn:
-            return ServiceUnknown('%s' % exn)
+        arcstat_result = await arcstat(query_jobids,
+                timeout = self.opts.arcstat_timeout, log = self.log)
+        if arcstat_result.is_error():
+            exn = arcstat_result.get_error()
+            return ServiceUnknown("Failed to query status of jobs: %s" % exn)
+        arcstat_response = arcstat_result.get()
         self.log.info('Queried %d jobs, found %d.',
-                      len(query_jobids), len(jobstat_by_id))
+                      len(query_jobids), len(arcstat_response.jobs))
 
         # Process jobs.
         job_action_stats = JobActionStats()
         futures = []
         for jobid in query_jobids:
             jobinfo = jobinfo_by_id[jobid]
-            if jobid in jobstat_by_id:
-                jobstat = jobstat_by_id[jobid]
-                future = \
-                    self.process_found_job(jobinfo, jobstat, job_action_stats)
-                next(future)
-                futures.append(future)
+            jobstat = arcstat_response.jobs.get(jobid)
+            if jobstat:
+                futures.append(asyncio.ensure_future(
+                    self.process_found_job(jobinfo, jobstat, job_action_stats)))
             else:
-                self.process_missing_job(jobinfo, job_action_stats)
-        for future in futures:
-            for _ in future:
-                pass
+                futures.append(asyncio.ensure_future(
+                    self.process_missing_job(jobinfo, job_action_stats)))
+        if futures != []:
+            await asyncio.gather(*futures)
         error_count += job_action_stats.error_count
 
         # Summary and report.
@@ -493,3 +495,6 @@ class Check_arcce_monitor(JobNagiosPlugin):
             self.log.info('- %s: %s', jobinfo.host_and_tag, jobinfo.job_state)
 
         return ServiceReport(status_code, msg)
+
+    def check(self):
+        return asyncio.run(self._check_async())
