@@ -6,12 +6,12 @@ from arcnagios.nagutils import ServiceUnknown
 from arcnagios.utils import ident, unspecified
 
 def boolean(s):
-    w = s.lower()
-    if w in ['0', 'false', 'no', 'off']:
+    s_lc = s.lower()
+    if s_lc in ['0', 'false', 'no', 'off']:
         return False
-    if w in ['1', 'true', 'yes', 'on']:
+    if s_lc in ['1', 'true', 'yes', 'on']:
         return True
-    raise ValueError('invalid literal for boolean: %r'%s)
+    raise ValueError('invalid literal for boolean: %r' % s)
 
 _interp_re = re.compile(r'%\(([a-zA-Z0-9_]+)\)')
 
@@ -54,16 +54,16 @@ class JobPlugin:
         return self.config.has_option(self.config_section, var)
 
     def getconf(self, var, default = unspecified, typ = ident, **kwargs):
-        if not default is unspecified and not self.hasconf(var):
+        if default is not unspecified and not self.hasconf(var):
             return default
         self._import_interpolations(var)
         self._update_vars(kwargs)
         try:
             return typ(self.config.get(self.config_section, var, **kwargs))
         except ValueError as exn:
-            raise ServiceUnknown('Bad value for configuration parameter %s in '
-                                 'section %s: %s'
-                                 % (var, self.config_section, exn))
+            raise ServiceUnknown(
+                    'Bad value for configuration parameter %s in section %s: %s'
+                    % (var, self.config_section, exn)) from exn
 
     def getconf_int(self, var, default = unspecified, **kwargs):
         self.getconf(var, default = default, type = int, **kwargs)
@@ -75,7 +75,7 @@ class JobPlugin:
         self.getconf(var, default = default, type = float, **kwargs)
 
     def getconf_strlist(self, var, default = unspecified, sep = None, **kwargs):
-        if not default is unspecified and not self.hasconf(var):
+        if default is not unspecified and not self.hasconf(var):
             return default
         self._import_interpolations(var)
         self._update_vars(kwargs)
