@@ -1,16 +1,22 @@
 """Base class for ARC monitoring and utilities for state tracking."""
 
+import asyncio
+from functools import cached_property
 from glob import glob
 import os
 import shutil
 import time
+from typing import Dict, List, Optional, Set, Tuple
 import jinja2
-from arcnagios import nagutils, persistence, vomsutils
-from arcnagios.arcutils import jobstate_of_str, ArcClient
+from arcnagios import arcclients, nagutils, persistence, vomsutils
+from arcnagios.arcutils import jobstate_of_str, ArcClient, JobState, ParseError
+from arcnagios.nagutils import NagiosPerflogTimer
+from arcnagios.persistence import ObsoletePersistentObject
 from arcnagios.reputation import ReputationTracker
 from arcnagios.rescheduler import Rescheduler
+from arcnagios.utils import host_of_uri
+from arcnagios.ce.jobplugin import JobPlugin
 from arcnagios.ce.jobplugins import load_jobplugin
-from arcnagios.utils import lazy_property
 
 class JobDescription:
     _required_attributes = ["job_name", "application_name", "script_path"]
@@ -50,6 +56,7 @@ class JobInfo(persistence.PersistentObject):
 
     persistence_version = 1
 
+    # TODO: Check if we can use introspection of the below declarations instead.
     persistent_attributes = {
         'submission_time':      persistence.pt_float,
         'host':                 persistence.pt_str,
@@ -69,23 +76,39 @@ class JobInfo(persistence.PersistentObject):
         'reputation_choices':   persistence.pt_json_opt,
     }
 
+    submission_time: float
+    host: str
+    job_tag: Optional[str]
+    progress_service: Optional[str]
+    termination_service: Optional[str]
+    job_id: str
+    job_state: JobState
+    job_specific_state: Optional[str]
+    job_state_time: Optional[float]
+    job_state_alert: Optional[int]
+    check_time: Optional[float]
+    check_attempts: Optional[int]
+    fetch_attempts: Optional[int]
+    stored_urls: List[str]
+    tests: List[str]
+    reputation_choices: Optional[Dict[str, str]]
+
     @property
     def host_and_tag(self):
         if self.job_tag:
             return '%s#%s' % (self.host, self.job_tag)
-        else:
-            return self.host
+        return self.host
 
     def __eq__(self, other):
         return self.host_and_tag == other.host_and_tag
     def __lt__(self, other):
         return self.host_and_tag < other.host_and_tag
 
-def key_value(s):
+def key_value(s: str) -> Tuple[str, str]:
     kv = s.split('=', 1)
     if len(kv) != 2:
         raise ValueError('Expecting an argument of the form KEY=VALUE.')
-    return kv
+    return (kv[0], kv[1])
 
 class JobNagiosPlugin(nagutils.NagiosPlugin, vomsutils.NagiosPluginVomsMixin):
     """Nagios probe to test ARC CEs.  The probe has two sub-commands
@@ -114,32 +137,38 @@ class JobNagiosPlugin(nagutils.NagiosPlugin, vomsutils.NagiosPluginVomsMixin):
     cleaner_arcclean_timeout = 5
     cleaner_arckill_timeout = 5
 
+    def _disconnect_reputation_tracker(self):
+        if self._reputation_tracker is not None:
+            self._reputation_tracker.disconnect()
+
     def __init__(self, **kwargs):
         default_template_dirs = glob(os.path.join(self.config_dir(),'*.d'))
         default_template_dirs.sort()
         nagutils.NagiosPlugin.__init__(self, **kwargs)
         self.arcclient = None
 
-        ap = self.argparser
-        ap.add_argument('--fqan', dest = 'fqan')
-        ap.add_argument('-O', dest = 'jobtest_options',
+        argp = self.argparser
+        argp.add_argument('--fqan', dest = 'fqan')
+        argp.add_argument('-O', dest = 'jobtest_options',
                 action = 'append', type = key_value, default = [],
                 help = 'Given a value of the form VAR=VALUE, binds VAR to '
                        'VALUE in the environment of the job tests.')
-        ap.add_argument('--template-dir', dest = 'template_dirs',
+        argp.add_argument('--template-dir', dest = 'template_dirs',
                 action = 'append', default = default_template_dirs,
                 help = 'Add a directory from which job description templates '
                        'can be loaded.')
-        ap.add_argument('--granular-perfdata', dest = 'granular_perfdata',
+        argp.add_argument('--granular-perfdata', dest = 'granular_perfdata',
                 default = False, action = 'store_true',
                 help = 'Report ARC command timing performance data per host '
                        'using labels of the form ARCCMD[HOST]. By default '
                        'report the aggretate time across hosts.')
+        argp.add_argument('--concurrency', type = int, default = 20,
+                help = 'The maximum number of ARC client processes to run '
+                        'concurrently.')
         self._reputation_tracker = None
-        # pylint: disable=unnecessary-lambda
-        self.at_exit(lambda: self._reputation_tracker.disconnect())
+        self.at_exit(self._disconnect_reputation_tracker)
 
-    def parse_args(self, args):
+    def parse_args(self, args: List[str]) -> None:
         nagutils.NagiosPlugin.parse_args(self, args)
         self.opts.template_dirs.reverse()
         self.arcclient = ArcClient(self.perflog)
@@ -147,32 +176,37 @@ class JobNagiosPlugin(nagutils.NagiosPlugin, vomsutils.NagiosPluginVomsMixin):
                 self.config,
                 os.path.join(self.opts.arcnagios_spooldir, 'reputation.db'))
 
-    def top_vopath(self, suffix):
+    @cached_property
+    def _arcclients_semaphore(self):
+        return asyncio.Semaphore(self.opts.concurrency or 20)
+
+    def top_vopath(self, suffix: str) -> str:
         return os.path.join(self.opts.arcnagios_spooldir,
                             self.voms_suffixed('ce') + '-' + suffix)
 
-    @lazy_property
-    def top_workdir(self):
+    @cached_property
+    def top_workdir(self) -> str:
         return self.top_vopath('state')
 
-    def workdir_for(self, host, job_tag):
+    def workdir_for(self, host: str, job_tag: Optional[str]):
         if job_tag:
             return os.path.join(self.top_workdir, host + '#' + job_tag)
         else:
             return os.path.join(self.top_workdir, host)
 
-    def archivedir_for(self, host, job_tag):
+    def archivedir_for(self, host: str, job_tag: Optional[str]):
         return os.path.join(self.top_vopath('archive'),
                             time.strftime('%Y-%m/%Y-%m-%d/%H:%M:%S-')
                             + host + (job_tag and ('#' + job_tag) or ''))
 
-    @lazy_property
-    def template_environment(self):
+    @cached_property
+    def template_environment(self) -> jinja2.Environment:
         return jinja2.Environment(
             loader=jinja2.FileSystemLoader(self.opts.template_dirs),
             autoescape=False)
 
-    def write_job_description(self, out_path, jobdesc):
+    def write_job_description(self, out_path: str, jobdesc: JobDescription) \
+            -> None:
         jobdesc.verify()
         try:
             templ = self.template_environment.get_template(jobdesc.template)
@@ -180,44 +214,59 @@ class JobNagiosPlugin(nagutils.NagiosPlugin, vomsutils.NagiosPluginVomsMixin):
             raise nagutils.ServiceUnknown('%s. The searched included %s.'
                     % (exn, ', '.join(self.opts.template_dirs)))
         content = templ.render(jd=jobdesc)
-        fd = open(out_path, 'w')
-        fd.write(content)
-        fd.close()
+        with open(out_path, 'w', encoding='utf-8') as fh:
+            fh.write(content)
 
-    def _cleaner_arcrm(self, url, n_attempts):
-        r = self.arcclient.arcrm(url, force = n_attempts > 8,
-                                 timeout = self.cleaner_arcrm_timeout)
-        if r.is_ok():
+    async def _cleaner_arcrm(self, url: str, n_attempts: int) -> bool:
+        se_host = host_of_uri(url)
+        with NagiosPerflogTimer(self.perflog, "arcrm_time", se_host):
+            result = await arcclients.arcrm(
+                    url, force = n_attempts > 8,
+                    timeout = self.cleaner_arcrm_timeout,
+                    log = self.log)
+        if result.is_ok():
             self.log.info('Removed test file %s.', url)
             return True
         else:
             self.log.warning('Failed to remove %s.', url)
             return False
 
-    def _cleaner_arcclean(self, job_id, n_attempts):
-        r = self.arcclient.arcclean(job_id, force = n_attempts > 8,
-                                    timeout = self.cleaner_arcclean_timeout)
-        if r.is_ok():
+    async def _cleaner_arcclean(self, job_id: str, n_attempts: int) -> bool:
+        ce_host = host_of_uri(job_id)
+        with NagiosPerflogTimer(self.perflog, "arcrm_time", ce_host):
+            result = await arcclients.arcclean(
+                    job_id, force = n_attempts > 8,
+                    timeout = self.cleaner_arcclean_timeout,
+                    log = self.log)
+        if result.is_ok():
             self.log.info('Removed job %s', job_id)
             return True
         else:
             self.log.warning('Failed to clean %s', job_id)
             return False
 
-    def _cleaner_arckill(self, job_id, n_attempts):
-        r = self.arcclient.arckill(job_id,
-                                   timeout = self.cleaner_arckill_timeout)
-        if r.is_ok(): return True
+    async def _cleaner_arckill(self, job_id: str, n_attempts: int) -> bool:
+        ce_host = host_of_uri(job_id)
+        with NagiosPerflogTimer(self.perflog, "arckill_time", ce_host):
+            result = await arcclients.arckill(
+                    job_id, timeout = self.cleaner_arckill_timeout,
+                    log = self.log)
+        if result.is_ok():
+            return True
 
-        r = self.arcclient.arcclean(job_id, force =  n_attempts > 8,
-                                    timeout = self.cleaner_arcclean_timeout)
-        if r.is_ok(): return True
+        result = await arcclients.arcclean(
+                job_id, force =  n_attempts > 8,
+                timeout = self.cleaner_arcclean_timeout,
+                log = self.log)
+        if result.is_ok():
+            self.log.info('Killed %s', job_id)
+            return True
+        else:
+            self.log.warning('Failed to kill %s', job_id)
+            return False
 
-        self.log.warning('Failed to kill %s: %s', job_id, r.error)
-        return False
-
-    @lazy_property
-    def cleaner(self):
+    @cached_property
+    def cleaner(self) -> Rescheduler:
         cleaner = Rescheduler(self.top_vopath('state.db'), 'cleaner',
                               nagios_report = self.nagios_report)
         cleaner.register('arcrm', self._cleaner_arcrm)
@@ -226,7 +275,8 @@ class JobNagiosPlugin(nagutils.NagiosPlugin, vomsutils.NagiosPluginVomsMixin):
         self.at_exit(cleaner.close)
         return cleaner
 
-    def cleanup_job_files(self, host, job_tag, archive = False):
+    def cleanup_job_files(self, host: str, job_tag: Optional[str],
+                          archive: bool = False) -> None:
         self.log.debug('Cleaning up job files for %s.', host)
         workdir = self.workdir_for(host, job_tag)
 #       archdir = os.path.join(workdir,
@@ -262,7 +312,8 @@ class JobNagiosPlugin(nagutils.NagiosPlugin, vomsutils.NagiosPluginVomsMixin):
         except OSError as exn:
             self.log.warning('Error clearing %s: %s', job_output_dir, exn)
 
-    def load_active_job(self, host, job_tag):
+    def load_active_job(self, host: str, job_tag: Optional[str]) \
+            -> Optional[JobInfo]:
         """Load information about the current job on `host : str` tagged with
         `job_tag : str`, or `None` if no information is found."""
 
@@ -271,14 +322,21 @@ class JobNagiosPlugin(nagutils.NagiosPlugin, vomsutils.NagiosPluginVomsMixin):
         if os.path.exists(ajf):
             self.log.debug('Loading job info from %s.', ajf)
             jobinfo = JobInfo()
-            jobinfo.persistent_load(
-                ajf, log=self.log,
-                persistence_version=jobinfo.persistence_version)
-            return jobinfo
-        else:
-            return None
+            try:
+                jobinfo.persistent_load(
+                    ajf, log=self.log,
+                    persistence_version=jobinfo.persistence_version)
+                return jobinfo
+            except ObsoletePersistentObject:
+                self.log.warning('Ignoring outdated job file %s.')
+                return None
+            except (OSError, ParseError) as exn:
+                self.log.error('Cannot load job file %s: %s', ajf, exn)
+                return None
+        return None
 
-    def save_active_job(self, jobinfo, host, job_tag):
+    def save_active_job(self, jobinfo: JobInfo,
+                        host: str, job_tag: Optional[str]) -> None:
         """Save information about the current job running on `host : str`
         tagged with `job_tag : str`."""
 
@@ -289,22 +347,22 @@ class JobNagiosPlugin(nagutils.NagiosPlugin, vomsutils.NagiosPluginVomsMixin):
             ajf, log=self.log,
             persistence_version=jobinfo.persistence_version)
 
-    def collect_active_jobids(self):
+    def collect_active_jobids(self) -> Set[str]:
         jobids = set()
-        for fp in glob(os.path.join(self.top_workdir, '*', 'active.jobid')):
+        for path in glob(os.path.join(self.top_workdir, '*', 'active.jobid')):
             try:
-                fh = open(fp)
-                jobids.add(fh.readline().strip())
+                with open(path, encoding='utf-8') as fh:
+                    jobids.add(fh.readline().strip())
             except IOError as exn:
-                self.log.error('Failed to read %s: %s', fp, exn)
+                self.log.error('Failed to read %s: %s', path, exn)
         return jobids
 
-    def discard_stored_urls(self, jobinfo):
+    async def _discard_stored_urls(self, jobinfo: JobInfo) -> None:
         for url in jobinfo.stored_urls:
             if not url.startswith('file:'):
-                self.cleaner.call('arcrm', url)
+                await self.cleaner.call('arcrm', url)
 
-    def cleanup_job_tests(self, jobinfo):
+    def cleanup_job_tests(self, jobinfo: JobInfo) -> None:
         for test_name in jobinfo.tests:
             try:
                 test = self.load_jobtest(test_name, hostname = jobinfo.host)
@@ -313,23 +371,27 @@ class JobNagiosPlugin(nagutils.NagiosPlugin, vomsutils.NagiosPluginVomsMixin):
                 self.log.error('Error in cleanup %s for %s: %s',
                                test_name, jobinfo.job_id, exn)
 
-    def cleanup_job(self, jobinfo, archive = False):
+    async def cleanup_job(
+            self, jobinfo: JobInfo, archive: bool = False
+        ) -> None:
         """Clean up job state from a fetched job."""
 
-        self.discard_stored_urls(jobinfo)
+        await self._discard_stored_urls(jobinfo)
         self.cleanup_job_tests(jobinfo)
         self.cleanup_job_files(jobinfo.host, jobinfo.job_tag, archive = archive)
 
-    def discard_job(self, jobinfo, archive = False):
+    async def discard_job(
+            self, jobinfo: JobInfo, archive: bool = False
+        ) -> None:
         """Discard the job described by `jobinfo : JobInfo`."""
 
         if jobinfo.job_state.is_final():
-            self.cleaner.call('arcclean', jobinfo.job_id)
+            await self.cleaner.call('arcclean', jobinfo.job_id)
         else:
-            self.cleaner.call('arckill', jobinfo.job_id)
-        self.cleanup_job(jobinfo, archive = archive)
+            await self.cleaner.call('arckill', jobinfo.job_id)
+        await self.cleanup_job(jobinfo, archive = archive)
 
-    def load_jobtest(self, jobtest_name, **env):
+    def load_jobtest(self, jobtest_name: str, **env) -> JobPlugin:
         """Load a plugin-based job-test from the section of the configuration
         specified by `jobtest_name`.  The result is an instance of `JobPlugin`
         subclass specified by the ``jobplugin`` variable of the given
