@@ -1,6 +1,8 @@
 """Combined argparse and ConfigParser."""
 
 import argparse
+from configparser import ConfigParser
+from typing import NoReturn, List, Union
 
 class UsageError(RuntimeError):
     pass
@@ -10,16 +12,20 @@ class ConfigError(RuntimeError):
 
 class ConfigArgumentParser(argparse.ArgumentParser):
 
-    def error(self, message):
+    def error(self, message: str) -> NoReturn:
         raise UsageError(message)
 
-    def configure_defaults(self, config, defaults_section):
+    def configure_defaults(
+            self, config: ConfigParser,
+            defaults_section: Union[List[str], str, None]) -> None:
         """Merge those configurations files in the list `config_paths` which
         exist and return the corresponding `ConfigParser` object.  Later files
         take presedence over former.  Variables from `defaults_section` will be
         used as defaults for the argument parser options."""
 
-        if not isinstance(defaults_section, list):
+        if not defaults_section:
+            defaults_section = []
+        elif not isinstance(defaults_section, list):
             defaults_section = [defaults_section]
         for section in defaults_section:
             if config.has_section(section):
@@ -36,4 +42,4 @@ class ConfigArgumentParser(argparse.ArgumentParser):
         except argparse.ArgumentTypeError as exn:
             raise ConfigError(
                     'The configuration corresponding to a command-line '
-                    'option failed: %s' % exn)
+                    'option failed: %s' % exn) from exn
